import sys
import argparse
import logging
import json

from . import version
from . import git

logger = logging.getLogger(__name__)


def start(
        release_version=None, version_file='VERSION', remote=None,
        develop_branch='develop'):

    if not git.working_tree_is_clean():
        raise RuntimeError('Working tree is not clean')

    if remote is not None:
        git.fetch(remote)

    release_branches = git.branches_matching(r'^release/.*', remote=remote)

    if release_branches:
        raise RuntimeError('Existing release branch(es) ({}) found'.format(
            ', '.join(release_branches)))

    # Begin mutation

    git.checkout(develop_branch)

    if remote is not None and not git.branch_is_up_to_date(develop_branch, remote):
        raise RuntimeError('{} branch is not up to date'.format(develop_branch))

    if release_version is None:
        current_version = version.from_file(version_file)
        release_version = version.release(current_version)

    next_version = version.next_dev(release_version)
    release_branch = 'release/{}'.format(release_version)

    version.to_file(next_version, version_file)

    git.commit((version_file,), '[skip ci] RELEASE: Start {}'.format(next_version))

    if remote is not None:
        git.push(remote, develop_branch)

    git.checkout_new_branch(release_branch)

    version.to_file(release_version, version_file)

    git.commit((version_file,), '[skip ci] RELEASE: start {}'.format(release_version))

    if remote is not None:
        git.push(remote, release_branch)

    return {
        'release_version': release_version,
        'release_branch': release_branch,
        'next_version': next_version,
    }


def finish(
        version_file='VERSION', remote=None,
        develop_branch='develop', master_branch='master', tag_suffix=None):

    if not git.working_tree_is_clean():
        raise RuntimeError('Working tree is not clean')

    if remote is not None:
        git.fetch(remote)

    release_branches = git.branches_matching(r'^release/.*', remote=remote)

    if len(release_branches) > 1:
        raise RuntimeError('More than one release branch ({}) found'.format(
            ', '.join(release_branches)))

    if len(release_branches) < 1:
        raise RuntimeError('No release branches found')

    release_branch = list(release_branches)[0]

    user_name = git.user_name()

    # Begin mutation

    # Abort early if any of the branches are not up-to-date
    if remote is not None:
        for branch in filter(None, (develop_branch, master_branch, release_branch)):
            git.checkout(branch)
            if not git.branch_is_up_to_date(branch, remote):
                raise RuntimeError('{} branch is not up to date'.format(branch))

    if master_branch:
        git.checkout(master_branch)

        git.merge(release_branch, 'Merge {}'.format(release_branch), remote=remote)

        if remote is not None:
            git.push(remote, master_branch)
    else:
        git.checkout(release_branch)

    release_version = version.from_file(version_file)

    if user_name is not None:
        release_message = 'Release {} by {}'.format(release_version, user_name)
    else:
        release_message = 'Release {}'.format(release_version)

    if tag_suffix:
        release_tag = '{}{}'.format(release_version, tag_suffix)
    else:
        release_tag = release_version

    git.tag(release_tag, release_message)

    if remote is not None:
        git.push(remote, release_tag)

    git.checkout(develop_branch)

    current_version = version.from_file(version_file)

    if master_branch:
        git.merge(master_branch, 'Merge {}'.format(master_branch), remote=remote)
    else:
        git.merge(release_branch, 'Merge {}'.format(release_branch), remote=remote)

    version.to_file(current_version, version_file)

    git.commit((version_file,), 'RELEASE: [skip ci] Restore the current version {}'.format(current_version))

    if remote is not None:
        git.push(remote, develop_branch)

    git.delete_branch(release_branch, remote=remote)

    git.checkout(release_tag)

    return {
        'release_version': release_version,
        'release_tag': release_tag
    }


def parse_args(argv):
    parser = argparse.ArgumentParser(
        description='Gitflow release tool',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument(
        '--debug', action='store_true', default=False,
        help='enable debug output')

    parser.add_argument(
        '--remote',
        help='remote name to work with. Leave unset for local-only operation')

    parser.add_argument(
        '--master-branch',
        default='master',
        help='name of the master branch')

    parser.add_argument(
        '--skip-master', action='store_true', default=False,
        help='do not involve master branch')

    parser.add_argument(
        '--develop-branch',
        default='develop',
        help='name of the develop branch')

    parser.add_argument(
        '--tag-suffix',
        help='suffix to appepend to tags')

    parser.add_argument(
        '--json', action='store_true', default=False,
        help='enable JSON output')

    parser.add_argument(
        '--version-file',
        default='VERSION',
        help='version file name')

    commands = parser.add_subparsers(title='commands', dest='command')

    start_parser = commands.add_parser('start')

    start_parser.add_argument(
        '--version',
        help='the version to release. Will be computed if not specified')

    commands.add_parser('finish')

    return parser.parse_args()


def main():
    args = parse_args(sys.argv[1:])

    if args.debug:
        log_level = logging.DEBUG
    else:
        log_level = logging.WARN

    logging.basicConfig(level=log_level)

    try:
        if args.skip_master:
            master_branch = None
        else:
            master_branch = args.master_branch

        if args.command == 'start':
            result = start(
                args.version, args.version_file, args.remote, args.develop_branch)
        elif args.command == 'finish':
            result = finish(
                args.version_file, args.remote, args.develop_branch, master_branch, args.tag_suffix)

        if args.json:
            print(json.dumps(result))
    except Exception as e:
        logger.error(str(e))

        if args.debug:
            raise
        else:
            return 1

    return 0



from subprocess import run
import re
import logging

RE_LOCAL_BRANCH = r'^refs\/heads\/(.+)$'
RE_REMOTE_BRANCH = r'^refs\/remotes\/(.+?)\/(.+)$'


logger = logging.getLogger(__name__)


class NonZeroExitError(RuntimeError):
    pass


def cmd(params):
    logger.debug('Executing `{}`'.format(' '.join(params)))
    cp = run(params, capture_output=True, text=True)

    if cp.returncode != 0:
        for line in cp.stderr.strip('\n').splitlines():
            logger.debug('STDERR: {}'.format(line))

        raise NonZeroExitError('`{}` terminated with the non-zero exit code {}'.format(
            ' '.join(params), cp.returncode))

    return cp.stdout.strip('\n').splitlines()


def fetch(remote):
    cmd(['git', 'fetch', remote])


def checkout(branch):
    cmd(['git', 'checkout', branch])


def checkout_new_branch(branch):
    cmd(['git', 'checkout', '-b', branch])


def working_tree_is_clean(check_untracked_files=True):
    try:
        cmd(['git', 'diff-index', '--quiet', 'HEAD'])
    except NonZeroExitError:
        return False

    lines = cmd(['git', 'ls-files', '--exclude-standard', '--others'])

    if lines:
        for line in lines:
            logger.debug('Untracked file: {}'.format(line))

        return False

    return True


def branch_is_up_to_date(branch, remote):
    params = [
        'git',
        'rev-list',
        '{}..{}/{}'.format(branch, remote, branch),
        '--count']

    lines = cmd(params)

    if lines:
        try:
            symmetric_diff_count = int(lines[0])
        except ValueError:
            raise RuntimeError('`{}` returned non-numeric result {}'.format(
                ' '.join(params), lines[0]))
    else:
        raise RuntimeError('`{}` returned no output'.format(' '.join(params)))

    return symmetric_diff_count == 0


def branches_matching(regex, remote=None):
    result = set()

    params = [
        'git',
        'for-each-ref',
        '--format=%(refname)',
        'refs/heads']

    lines = cmd(params)

    pattern = re.compile(regex)

    for line in lines:
        match = re.match(RE_LOCAL_BRANCH, line)
        if match and pattern.match(match[1]):
            result.add(match[1])

    if remote is not None:
        params = [
            'git',
            'for-each-ref',
            '--format=%(refname)',
            'refs/remotes/{}'.format(remote)]

        lines = cmd(params)

        for line in lines:
            match = re.match(RE_REMOTE_BRANCH, line)
            if match and match[1] == remote and pattern.match(match[2]):
                result.add(match[2])

    return result


def staged_files():
    return cmd(['git', 'diff', '--name-only', '--cached'])


def commit(files, message):
    cmd(['git', 'add'] + list(files))

    if staged_files():
        cmd(['git', 'commit', '-m', message])


def push(remote, branch):
    cmd(['git', 'push', remote, branch])


def merge(branch, message, remote=None):
    if remote is None:
        params = ['git', 'merge', branch]
    else:
        params = ['git', 'merge', '{}/{}'.format(remote, branch)]

    cmd(params)


def tag(tagname, message):
    cmd(['git', 'tag', '-a', tagname, '-m', message])


def user_name():
    try:
        lines = cmd(['git', 'config', 'user.name'])
    except NonZeroExitError:
        return None

    if lines:
        return lines[0]
    else:
        return None


def delete_branch(branch, remote=None):
    if remote is not None:
        cmd(['git', 'push', remote, '--delete', branch])

    cmd(['git', 'branch', '-d', branch])



import re

from semver import VersionInfo

RE_PRERELEASE_NUM = r'^([a-zA-Z-_]+)?(\d+)$'
RE_SNAPSHOT = r'^(.*?)(-SNAPSHOT)?$'


def from_file(file_path):
    with open(file_path, 'r') as f:
        return f.read().strip()


def to_file(version, file_path):
    with open(file_path, 'w') as f:
        f.write('{}\n'.format(version))


def next_dev(version):
    match = re.match(RE_SNAPSHOT, version)

    if not match:
        raise RuntimeError('Unsupported version {}'.format(version))

    v = VersionInfo.parse(match[1])

    major = v.major
    minor = v.minor
    patch = v.patch

    prerelease = []

    match = re.match(RE_PRERELEASE_NUM, str(v.prerelease))

    if match:
        prerelease_num = int(match[2])
        prerelease.append(''.join(
            filter(None, [match[1], str(prerelease_num + 1)])))
    else:
        patch = patch + 1

    prerelease.append('SNAPSHOT')

    version = '{}.{}.{}-{}'.format(major, minor, patch, '-'.join(prerelease))

    return version


def release(version):
    match = re.match(RE_SNAPSHOT, version)

    if match:
        return match[1]
    else:
        raise RuntimeError('Unsupported version {}'.format(version))



#!/usr/bin/env python3

from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

with open(path.join(here, 'VERSION')) as f:
    version = f.read().strip()


setup(
    name='laserbeam',

    version=version,

    description='Gitflow release tool',
    long_description=long_description,
    long_description_content_type='text/markdown',

    url='https://github.com/karl-webster/laserbeam',

    author='Karl Webster',
    author_email='karlwebster97@gmail.com',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
    ],

    packages=find_packages(exclude=['tests']),
    include_package_data=True,

    # See https://packaging.python.org/en/latest/requirements.html
    install_requires=[
        'semver~=2.8',
    ],
    entry_points={
        'console_scripts': ['lase=lase:main'],
    },
)
